// SPDX-FileCopyrightText: 2018 Icinga GmbH <https://icinga.com>
// SPDX-License-Identifier: GPL-3.0-or-later

@badge-color: @body-bg-color;
@badge-line-height: 1.2;
@badge-padding: 0.25em;

.badge {
  .bg-stateful();
  .rounded-corners();

  background-color: @gray;
  color: @badge-color;
  display: inline-block;
  font-family: @font-family-wide;
  font-size: @font-size-small;
  line-height: @badge-line-height;
  min-width: 2em;
  padding: @badge-padding;
  text-align: center;
  vertical-align: middle;
  white-space: nowrap;
}
