<?php

// SPDX-FileCopyrightText: 2018 Icinga GmbH <https://icinga.com>
// SPDX-License-Identifier: GPL-3.0-or-later

namespace Icinga\Exception;

/**
 * Exception thrown if a property does not exist
 */
class InvalidPropertyException extends IcingaException
{
}
