// SPDX-FileCopyrightText: 2018 Icinga GmbH <https://icinga.com>
// SPDX-License-Identifier: GPL-3.0-or-later

@colorMainLayout: @icinga-blue;
@colorMainBackground: @body-bg-color;
@colorMainForeground: @text-color;
@colorMainLink: @text-color;
@colorSecondary: @gray-lightest;
@colorGray: @gray-lightest;
@colorLinkDefault: @text-color;
@colorTextDefault: @text-color;
@colorTextDarkDefault: @text-color;
@colorOk: @color-ok;
@colorWarning: #ffaa44;
@colorWarningHandled: #ffcc66;
@colorCritical: #ff5566;
@colorCriticalHandled: #ff99aa;
@colorUnknown: #aa44ff;
@colorUnknownHandled: #cc77ff;
@colorUnreachable: #aa44ff;
@colorUnreachableHandled: #cc77ff;
@colorPending: #77aaff;
@colorInvalid: #999;
@colorFormNotificationInfo: #77aaff;
@colorFormNotificationWarning: #ffaa44;
@colorFormNotificationError: #ff5566;
@colorPetrol: @icinga-blue;
@menu-2ndlvl-highlight-color: @menu-2ndlvl-active-color;

table.action {
  .common-table();
}

table.avp {
  .name-value-table();
}

input.search {
  background-size: 0;
}
