<div class="controls">
<?php if (! $this->compact): ?>
<?= $this->tabs ?>
<?php endif ?>
</div>
<?php if ($this->dashboard): ?>
    <div class="dashboard content">
    <?= $this->dashboard ?>
    </div>
<?php else: ?>
    <div class="content">
        <h1><?= $this->escape($this->translate('Welcome to Icinga Web!')) ?></h1>
        <p>
        <?php if (! $this->hasPermission('config/modules')) {
            echo $this->escape($this->translate(
                'Currently there is no dashlet available. Please contact the administrator.'
            ));
        } else {
            printf(
                $this->escape($this->translate('Currently there is no dashlet available. This might change once you enabled some of the available %s.')),
                $this->qlink($this->translate('modules'), 'config/modules')
            );
        } ?>
        </p>
    </div>
<?php endif ?>
