// SPDX-FileCopyrightText: 2019 Icinga GmbH <https://icinga.com>
// SPDX-License-Identifier: GPL-3.0-or-later

@color-ok: fade(#77E08E, 25%);
@color-critical: #FE5566;
@color-critical-handled: fade(@color-critical, 33%);
@color-warning: #B0A029;
@color-warning-handled: fade(@color-warning, 33%);
@color-unknown: #7791E0;
@color-unknown-handled: fade(@color-unknown, 50%);
@color-unreachable: @color-unknown;
@color-unreachable-handled: @color-unknown-handled;
@color-pending: fade(#FFFFFF, 75%);

/* Adapt font color to match handled / unhandled states */
.badge,
.state-badge {
  font-weight: bold;
  color: @text-color !important;

  &.handled,
  &.state-up,
  &.state-ok {
    color: fade(@text-color, 75%) !important;
  }
}

.processinfo .process > div.backend-running {
  color: @text-color;
}
