<?php

// SPDX-FileCopyrightText: 2018 Icinga GmbH <https://icinga.com>
// SPDX-License-Identifier: GPL-3.0-or-later

namespace Icinga\Application\Logger\Writer;

/**
 * Deprecated, compat only.
 *
 * Use Icinga\Application\Logger\Writer\StderrWriter instead.
 */
class StdoutWriter extends StderrWriter
{
}
